import numpy as np
from math import ceil
from hapla import shared_cy
from hapla import admix_cy


##### hapla - functions #####
### hapla struct
# SVD through eigendecomposition
def eigSVD(H):
	D, V = np.linalg.eigh(np.dot(H.T, H))
	S = np.sqrt(D)
	U = np.dot(H, V*(1.0/S))
	return np.ascontiguousarray(U[:,::-1]), np.ascontiguousarray(S[::-1]), \
		np.ascontiguousarray(V[:,::-1])

# Randomized PCA with dynamic shift
def randomizedSVD(Z_agg, p_vec, a_vec, K, chunk, power, rng):
	M, N = Z_agg.shape
	W = ceil(M/chunk)
	a = 0.0
	L = max(K + 10, 20)
	H = np.zeros((N, L), dtype=np.float32)
	X = np.zeros((chunk, N), dtype=np.float32)
	A = rng.standard_normal(size=(M, L), dtype=np.float32)

	# Prime iteration
	for w in np.arange(W):
		M_w = w*chunk
		M_e = min((w + 1)*chunk, M)
		M_x = M_e - M_w
		shared_cy.chunkZ(Z_agg[M_w:M_e], X[:M_x], p_vec[M_w:M_e], a_vec[M_w:M_e])
		H += np.dot(X[:M_x].T, A[M_w:M_e])
	Q, _, _ = eigSVD(H)
	H.fill(0.0)

	# Power iterations
	for p in np.arange(power):
		print(f"\rPower iteration {p+1}/{power}", end="")
		for w in np.arange(W):
			M_w = w*chunk
			M_e = min((w + 1)*chunk, M)
			M_x = M_e - M_w
			shared_cy.chunkZ(Z_agg[M_w:M_e], X[:M_x], p_vec[M_w:M_e], a_vec[M_w:M_e])
			A[M_w:M_e] = np.dot(X[:M_x], Q)
			H += np.dot(X[:M_x].T, A[M_w:M_e])
		H -= a*Q
		Q, S, _ = eigSVD(H)
		H.fill(0.0)
		if S[-1] > a:
			a = 0.5*(S[-1] + a)

	# Extract singular vectors
	for w in np.arange(W):
		M_w = w*chunk
		M_e = min((w + 1)*chunk, M)
		M_x = M_e - M_w
		shared_cy.chunkZ(Z_agg[M_w:M_e], X[:M_x], p_vec[M_w:M_e], a_vec[M_w:M_e])
		A[M_w:M_e] = np.dot(X[:M_x], Q)
	U, S, V = eigSVD(A)
	return U[:,:K], S[:K], np.dot(Q, V)[:,:K]



### hapla admix
# Update for ancestry estimation
def steps(Z, P, Q, Q_tmp, k_vec, c_vec, y, L):
	admix_cy.updateP(Z, P, Q, Q_tmp, k_vec, c_vec, L)
	admix_cy.updateQ(Q, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q, y)

# Accelerated update for ancestry estimation
def quasi(Z, P0, Q0, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y, L):
	# 1st EM step
	admix_cy.accelP(Z, P0, P1, Q0, Q_tmp, k_vec, c_vec, L)
	admix_cy.accelQ(Q0, Q1, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q1, y)

	# 2nd EM step
	admix_cy.accelP(Z, P1, P2, Q1, Q_tmp, k_vec, c_vec, L)
	admix_cy.accelQ(Q1, Q2, Q_tmp, Z.shape[0])
	if y is not None:
		admix_cy.superQ(Q2, y)

	# Acceleation update
	admix_cy.jumpP(P0, P1, P2, k_vec, c_vec, Q0.shape[1])
	admix_cy.jumpQ(Q0, Q1, Q2)
	if y is not None:
		admix_cy.superQ(Q0, y)

# Batch accelerated update for ancestry estimation
def batQuasi(Z, P0, Q0, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, s_bat, y, L):
	# 1st EM step
	admix_cy.accelBatchP(Z, P0, P1, Q0, Q_tmp, k_vec, c_vec, s_bat, L)
	admix_cy.accelQ(Q0, Q1, Q_tmp, s_bat.shape[0])
	if y is not None:
		admix_cy.superQ(Q1, y)

	# 2nd EM step
	admix_cy.accelBatchP(Z, P1, P2, Q1, Q_tmp, k_vec, c_vec, s_bat, L)
	admix_cy.accelQ(Q1, Q2, Q_tmp, s_bat.shape[0])
	if y is not None:
		admix_cy.superQ(Q2, y)

	# Acceleation update
	admix_cy.jumpBatchP(P0, P1, P2, k_vec, c_vec, s_bat, Q0.shape[1])
	admix_cy.jumpQ(Q0, Q1, Q2)
	if y is not None:
		admix_cy.superQ(Q0, y)

# Randomized PCA with dynamic shift for ALS/SVD initialization
def centerSVD(Z, p_vec, k_vec, c_vec, W, K, chunk, power, rng):
	N = Z.shape[1]
	D = K - 1
	M = c_vec[W]
	B = ceil(chunk/ceil(M/W))
	C = ceil(W/B)
	a = 0.0
	L = max(D + 10, 20)
	H = np.zeros((N, L), dtype=np.float32)
	X = np.zeros((np.max(k_vec[:W])*B, N), dtype=np.float32)
	A = rng.standard_normal(size=(M, L), dtype=np.float32)

	# Prime iteration
	for c in np.arange(C):
		W_b = c*B
		W_e = min((c + 1)*B, W)
		C_b = c_vec[W_b]
		C_e = c_vec[W_e]
		C_x = C_e - C_b
		shared_cy.centerC(Z[W_b:W_e], X[:C_x], p_vec[C_b:C_e], k_vec[W_b:W_e], c_vec[W_b:W_e])
		H += np.dot(X[:C_x].T, A[C_b:C_e])
	Q, _, _ = eigSVD(H)
	H.fill(0.0)

	# Power iterations
	for _ in np.arange(power):
		for c in np.arange(C):
			W_b = c*B
			W_e = min((c + 1)*B, W)
			C_b = c_vec[W_b]
			C_e = c_vec[W_e]
			C_x = C_e - C_b
			shared_cy.centerC(Z[W_b:W_e], X[:C_x], p_vec[C_b:C_e], k_vec[W_b:W_e], c_vec[W_b:W_e])
			A[C_b:C_e] = np.dot(X[:C_x], Q)
			H += np.dot(X[:C_x].T, A[C_b:C_e])
		H -= a*Q
		Q, S, _ = eigSVD(H)
		H.fill(0.0)
		if S[-1] > a:
			a = 0.5*(S[-1] + a)

	# Extract singular vectors
	for c in np.arange(C):
		W_b = c*B
		W_e = min((c + 1)*B, W)
		C_b = c_vec[W_b]
		C_e = c_vec[W_e]
		C_x = C_e - C_b
		shared_cy.centerC(Z[W_b:W_e], X[:C_x], p_vec[C_b:C_e], k_vec[W_b:W_e], c_vec[W_b:W_e])
		A[C_b:C_e] = np.dot(X[:C_x], Q)
	U, S, V = eigSVD(A)
	U = np.ascontiguousarray(U[:,:D])
	S = np.ascontiguousarray(S[:D])
	V = np.ascontiguousarray(np.dot(Q, V)[:,:D])
	return U, S, V

# Alternating least square (ALS) for initializing Q and P
def factorALS(U, S, V, p_vec, k_vec, c_vec, iter, tole, rng):
	M, D = U.shape
	Y = np.ascontiguousarray(U*S)
	P = rng.random(size=(M, D + 1), dtype=np.float32)
	admix_cy.projectP(P, k_vec, c_vec)
	I = np.dot(P, np.linalg.pinv(np.dot(P.T, P)))
	Q = np.dot(V, np.dot(Y.T, I)) + np.sum(I*p_vec.reshape(-1,1), axis=0)
	admix_cy.projectQ(Q)
	Q0 = np.copy(Q)

	# Perform ALS iterations
	for _ in range(iter):
		# Update P
		I = np.dot(Q, np.linalg.pinv(np.dot(Q.T, Q)))
		P = np.dot(Y, np.dot(V.T, I)) + np.outer(p_vec, np.sum(I, axis=0))
		admix_cy.projectP(P, k_vec, c_vec)

		# Update Q
		I = np.dot(P, np.linalg.pinv(np.dot(P.T, P)))
		Q = np.dot(V, np.dot(Y.T, I)) + np.sum(I*p_vec.reshape(-1,1), axis=0)
		admix_cy.projectQ(Q)

		# Check convergence
		if admix_cy.rmseQ(Q, Q0) < tole:
			break
		memoryview(Q0.ravel())[:] = memoryview(Q.ravel())
	return P.flatten().astype(float), Q.astype(float)

# Project remaining samples for ALS/SVD initialization
def centerSub(Z, S, V, p_vec, k_vec, c_vec, W_sub, chunk):
	N = Z.shape[1]
	D = V.shape[1]
	W = Z.shape[0] - W_sub
	A = c_vec[W_sub]
	M = c_vec[Z.shape[0]] - A
	B = ceil(chunk/ceil(M/W))
	C = ceil(W/B)
	Y = np.ascontiguousarray(V*(1.0/S))
	U = np.zeros((M, D), dtype=np.float32)
	X = np.zeros((np.max(k_vec[W_sub:])*B, N), dtype=np.float32)

	# Loop through chunks
	for c in np.arange(C):
		W_b = W_sub + c*B
		W_e = W_sub + min((c + 1)*B, W)
		C_b = c_vec[W_b]
		C_e = c_vec[W_e]
		C_x = C_e - C_b
		shared_cy.centerC(Z[W_b:W_e], X[:C_x], p_vec[C_b:C_e], k_vec[W_b:W_e], c_vec[W_b:W_e])
		U[(C_b - A):(C_e - A)] = np.dot(X[:C_x], Y)
	return U

# Least square (ALS) for subsampled P and Q followed by standard iteration
def factorSub(U_sub, U_rem, S, V, p_vec, k_vec, c_vec, W_sub, iter, tole, rng):
	# Subsampled arrays
	M, D = U_sub.shape
	Y = np.ascontiguousarray(U_sub*S)
	p_sub = p_vec[:c_vec[W_sub]]
	k_sub = k_vec[:W_sub]
	c_sub = c_vec[:W_sub]

	# Initiate P and Q
	P = rng.random(size=(M, D + 1), dtype=np.float32)
	admix_cy.projectP(P, k_sub, c_sub)
	I = np.dot(P, np.linalg.pinv(np.dot(P.T, P)))
	Q = np.dot(V, np.dot(Y.T, I)) + np.sum(I*p_sub.reshape(-1,1), axis=0)
	admix_cy.projectQ(Q)
	Q0 = np.copy(Q)

	# Perform ALS iterations
	for _ in range(iter):
		# Update P
		I = np.dot(Q, np.linalg.pinv(np.dot(Q.T, Q)))
		P = np.dot(Y, np.dot(V.T, I)) + np.outer(p_sub, np.sum(I, axis=0))
		admix_cy.projectP(P, k_sub, c_sub)

		# Update Q
		I = np.dot(P, np.linalg.pinv(np.dot(P.T, P)))
		Q = np.dot(V, np.dot(Y.T, I)) + np.sum(I*p_sub.reshape(-1,1), axis=0)
		admix_cy.projectQ(Q)

		# Check convergence
		if admix_cy.rmseQ(Q, Q0) < tole:
			break
		memoryview(Q0.ravel())[:] = memoryview(Q.ravel())
	del Q0, p_sub, k_sub, c_sub

	# Perform extra full ALS iteration
	Y = np.ascontiguousarray(np.concatenate((U_sub, U_rem), axis=0)*S)
	I = np.dot(Q, np.linalg.pinv(np.dot(Q.T, Q)))
	P = np.dot(Y, np.dot(V.T, I)) + np.outer(p_vec, np.sum(I, axis=0))
	admix_cy.projectP(P, k_vec, c_vec)
	I = np.dot(P, np.linalg.pinv(np.dot(P.T, P)))
	Q = np.dot(V, np.dot(Y.T, I)) + np.sum(I*p_vec.reshape(-1,1), axis=0)
	admix_cy.projectQ(Q)
	return P.flatten().astype(float), Q.astype(float)

# Update for ancestry estimation in projection mode
def proSteps(Z, P, Q, Q_tmp, k_vec, c_vec):
	admix_cy.stepQ(Z, P, Q, Q_tmp, k_vec, c_vec)
	admix_cy.updateQ(Q, Q_tmp, Z.shape[0])

# Accelerated update for ancestry estimation in projection mode
def proQuasi(Z, P, Q0, Q_tmp, Q1, Q2, k_vec, c_vec):
	# 1st EM step
	admix_cy.stepQ(Z, P, Q0, Q_tmp, k_vec, c_vec)
	admix_cy.accelQ(Q0, Q1, Q_tmp, Z.shape[0])

	# 2nd EM step
	admix_cy.stepQ(Z, P, Q1, Q_tmp, k_vec, c_vec)
	admix_cy.accelQ(Q1, Q2, Q_tmp, Z.shape[0])

	# Acceleation update
	admix_cy.jumpQ(Q0, Q1, Q2)

# Batch accelerated update for ancestry estimation in projection mode
def proBatch(Z, P, Q0, Q_tmp, Q1, Q2, k_vec, c_vec, s_bat):
	# 1st EM step
	admix_cy.stepBatchQ(Z, P, Q0, Q_tmp, k_vec, c_vec, s_bat)
	admix_cy.accelQ(Q0, Q1, Q_tmp, s_bat.shape[0])

	# 2nd EM step
	admix_cy.stepBatchQ(Z, P, Q1, Q_tmp, k_vec, c_vec, s_bat)
	admix_cy.accelQ(Q1, Q2, Q_tmp, s_bat.shape[0])

	# Acceleation update
	admix_cy.jumpQ(Q0, Q1, Q2)



"""
hapla.
Ancestry estimation using accelerated EM algorithm.
"""

__author__ = "Jonas Meisner"

# Libraries
import os
from datetime import datetime
from time import time

VERSION = "0.31.0"

##### hapla admix #####
def main(args, deaf):
	print("-----------------------------------")
	print(f"hapla by Jonas Meisner (v{VERSION})")
	print(f"hapla admix using {args.threads} thread(s)")
	print("-----------------------------------")

	# Check input
	assert (args.filelist is not None) or (args.clusters is not None), "No input data (--filelist or --clusters)!"
	assert args.K > 1, "Please select K > 1!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.seed >= 0, "Please select a valid seed!"
	assert args.iter > 0, "Please select a valid number of iterations!"
	assert args.tole >= 0.0, "Please select a valid tolerance!"
	assert args.batches > 0, "Please select a valid number of mini-batches!"
	assert args.check > 0, "Please select a valid value for convergence check!"
	assert args.power > 0, "Please select a valid number of power iterations!"
	assert args.chunk > 0, "Please select a valid SVD chunk size!"
	assert args.als_iter > 0, "Please select a valid number of iterations in ALS!"
	assert args.als_tole >= 0.0, "Please select a valid tolerance in ALS!"
	assert args.subsampling > 1, "Please select a valid subsampling factor!"
	print("Estimating ancestry proportions.")
	print(f"K={args.K}, seed={args.seed}, batches={args.batches}\n")
	start = time()

	# Create log-file of used arguments
	full = vars(args)
	mand = ["seed", "batches"]
	with open(f"{args.out}.K{args.K}.s{args.seed}.log", "w") as log:
		log.write(f"hapla v{VERSION}\n")
		log.write("hapla admix\n")
		log.write(f"Time: {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}\n")
		log.write(f"Directory: {os.getcwd()}\n")
		log.write("Options:\n")
		for key in full:
			if full[key] != deaf[key]:
				if type(full[key]) is bool:
					log.write(f"\t--{key}\n")
				else:
					log.write(f"\t--{key} {full[key]}\n")
			elif key in mand:
				log.write(f"\t--{key} {full[key]}\n")
	del full, deaf, mand

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Import numerical libraries and cython functions
	import numpy as np
	from math import ceil
	from hapla import functions
	from hapla import admix_cy
	from hapla import shared_cy

	# Prepare list of data files
	if args.filelist is not None:
		W = 0 # Counter for windows
		Z_list = [] # List of filenames
		with open(args.filelist) as f:
			for z_file in f:
				# Check input across files (chromosomes) and count windows
				z = z_file.strip("\n")
				Z_list.append(z)
				assert os.path.isfile(f"{z}.bca"), "bca file doesn't exist!"
				assert os.path.isfile(f"{z}.ids"), "ids file doesn't exist!"
				assert os.path.isfile(f"{z}.win"), "win file doesn't exist!"
				if W == 0: # First file
					z_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					k_vec = np.genfromtxt(f"{z}.win", dtype=np.uint32, usecols=[5])
					N = z_ids.shape[0]
					W = k_vec.shape[0]
					w_list = [W]
				else: # Loop files
					t_ids = np.genfromtxt(f"{z}.ids", dtype=np.str_)
					assert np.sum(z_ids != t_ids) == 0, "Samples don't match across files!"
					k_tmp = np.genfromtxt(f"{z}.win", dtype=np.uint32, usecols=[5])
					k_vec = np.append(k_vec, k_tmp)
					W += k_tmp.shape[0]
					w_list.append(k_tmp.shape[0])
		F = len(Z_list)
		w_vec = np.array(w_list, dtype=np.uint32)
		f_vec = np.insert(np.cumsum(w_vec, dtype=np.uint32), 0, 0)
		del z_ids, t_ids, k_tmp, w_list
	else: # Single file (chromosome)
		F = 1
		Z_list = [args.clusters]
		assert os.path.isfile(f"{Z_list[0]}.bca"), "bca file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.ids"), "ids file doesn't exist!"
		assert os.path.isfile(f"{Z_list[0]}.win"), "win file doesn't exist!"
		k_vec = np.genfromtxt(f"{Z_list[0]}.win", dtype=np.uint32, usecols=[5])
		N = np.genfromtxt(f"{Z_list[0]}.ids", dtype=np.str_).shape[0]
		W = k_vec.shape[0]
		w_vec = np.array([W], dtype=np.uint32)
	print(f"Parsing {F} file(s).")

	# Load haplotype cluster assignments from binary hapla format
	B = 0
	Z = np.zeros((W, 2*N), dtype=np.uint8)
	for z in np.arange(F):
		with open(f"{Z_list[z]}.bca", "rb") as f:
			# Check magic numbers
			magic = np.fromfile(f, dtype=np.uint8, count=3)
			assert np.allclose(magic, np.array([7, 9, 13], dtype=np.uint8)), "Magic number doesn't match file format!"
			
			# Add haplotype cluster assignments to container
			z_tmp = np.fromfile(f, dtype=np.uint8)
			z_tmp.shape = (w_vec[z], 2*N)
			Z[B:(B + w_vec[z]),:] = z_tmp
			B += w_vec[z]
		print(f"\rParsed file {z + 1}/{F}", end="")
	del magic, z_tmp

	# Count haplotype cluster alleles
	M = np.sum(k_vec, dtype=np.uint32)
	L_nrm = 2.0*float(M)*float(N)
	c_vec = np.insert(np.cumsum(k_vec*args.K, dtype=np.uint32), 0, 0)

	# Print information
	print("\rLoaded haplotype cluster assignments:\n" + \
		f"- {N} samples\n" + \
		f"- {W} windows\n" + \
		f"- {M} clusters\n")

	# Set up parameters
	rng = np.random.default_rng(args.seed)
	if args.supervised is not None: # Supervised mode
		# Check input of ancestral sources
		print("Ancestry estimation in supervised mode!")
		assert os.path.isfile(args.supervised), "Population assignment file doesn't exist!"
		y = np.genfromtxt(args.supervised, dtype=np.uint8).reshape(-1)
		assert y.shape[0] == N, f"Number of samples differ between files!"
		assert np.max(y) <= args.K, "Wrong number of ancestral sources!"
		assert np.min(y) >= 0, "Wrong format for population assignments!"
		print(f"{np.sum(y > 0)}/{N} samples with fixed ancestry.")
		y = np.repeat(y, 2) # Repeat the vector for easy computations

		# Count ancestral sources
		x = np.unique(y[y > 0])
		x -= 1
		x = np.sort(x)

		# Initialize parameters
		P = rng.random(size=(M, args.K)).clip(min=1e-5, max=1-(1e-5))
		Q = rng.random(size=(2*N, args.K)).clip(min=1e-5, max=1-(1e-5))
		Q /= np.sum(Q, axis=1, keepdims=True)
		P[:,x] = 0.0
		c_tmp = np.insert(np.cumsum(k_vec, dtype=np.uint32), 0, 0)
		admix_cy.superP(Z, P, k_vec, c_tmp, y)
		admix_cy.superQ(Q, y)
		P = P.flatten()
		del x, c_tmp
	elif args.projection is not None: # Projection mode
		# Load ancestral haplotype cluster frequencies
		print("Ancestry estimation in projection mode!")
		assert os.path.isfile(args.projection), "P matrix file/filelist doesn't exist!"
		if F > 1: # Load multiple frequency files from filelist
			P_list = []
			with open(args.projection) as f:
				for p, p_file in enumerate(f):
					assert os.path.isfile(p_file.strip("\n")), f"The {p + 1}/{F} matrix file doesn't exist!"
					P_list.append(p_file.strip("\n"))
			assert len(P_list) == F, "Number of files doesn't match!"

			# Load in files to full matrix
			P = np.zeros(M*args.K)
			M_tmp = 0
			for p in np.arange(F):
				p_tmp = np.fromfile(P_list[p], dtype=float)
				M_tmp += p_tmp.shape[0]
				P[c_vec[f_vec[p]]:c_vec[f_vec[p + 1]]] = p_tmp
			assert np.any(P > 0.0), "Wrong format for haplotype cluster alleles!"
			assert M_tmp == (M*args.K), "Number of haplotype clusters doesn't match!"
			del P_list, p_tmp
		else: # Load single frequency file
			P = np.fromfile(args.projection, dtype=float)
			assert P.shape[0] == (M*args.K), "Number of haplotype clusters doesn't match!"
		
		# Check that cluster frequencies sum to one
		p_sum = np.zeros((W, args.K))
		admix_cy.checkP(P, p_sum, k_vec, c_vec, args.K)
		assert np.allclose(p_sum, 1.0), "Wrong format for haplotype cluster alleles!"
		del p_sum

		# Initialize Q matrix
		Q = rng.random(size=(2*N, args.K)).clip(min=1e-5, max=1-(1e-5))
		Q /= np.sum(Q, axis=1, keepdims=True)
	else:
		if args.random_init: # Random initialization
			print("Random initialization.")
			P = rng.random(size=(M*args.K)).clip(min=1e-5, max=1-(1e-5))
			Q = rng.random(size=(2*N, args.K)).clip(min=1e-5, max=1-(1e-5))
			Q /= np.sum(Q, axis=1, keepdims=True)
			admix_cy.createP(P, k_vec, c_vec, args.K)
		else: # SVD/ALS initialization
			print("SVD/ALS initialization.", end="", flush=True)
			ts = time()
			p_vec = np.zeros(M, dtype=np.float32)
			c_tmp = np.insert(np.cumsum(k_vec, dtype=np.uint32), 0, 0)
			shared_cy.estimateFreq(Z, p_vec, k_vec, c_tmp)
			if F > 1: # Initialize based on subsampled chromosomes/files
				W_s = f_vec[ceil(F/args.subsampling)]
				U_s, S, V = functions.centerSVD(Z, p_vec, k_vec, c_tmp, W_s, args.K, args.chunk, args.power, rng)
				U_r = functions.centerSub(Z, S, V, p_vec, k_vec, c_tmp, W_s, args.chunk)
				P, Q = functions.factorSub(U_s, U_r, S, V, p_vec, k_vec, c_tmp, W_s, args.als_iter, args.als_tole, rng)
				del U_s, U_r
			else:
				U, S, V = functions.centerSVD(Z, p_vec, k_vec, c_tmp, W, args.K, args.chunk, args.power, rng)
				P, Q = functions.factorALS(U, S, V, p_vec, k_vec, c_tmp, args.als_iter, args.als_tole, rng)
				del U
			print(f"\rSVD/ALS initialization.\t\t({time() - ts:.1f}s)")
			del p_vec, c_tmp, S, V
		y = None

	# Set up containers for EM algorithm
	Q1 = np.zeros_like(Q)
	Q2 = np.zeros_like(Q)
	Q_tmp = np.zeros_like(Q)
	if args.projection is None:
		L = np.max(k_vec) # Max clusters for inner dynamic arrays
		P1 = np.zeros_like(P)
		P2 = np.zeros_like(P)

	# Estimate initial log-likelihood
	L_pre = admix_cy.loglike(Z, P, Q, c_vec)
	print(f"Initial log-like: {L_pre*L_nrm:.1f}")

	# Accelerated priming iteration
	ts = time()
	print("Priming iteration.", end="", flush=True)
	if args.projection is None:
		functions.steps(Z, P, Q, Q_tmp, k_vec, c_vec, y, L)
		functions.quasi(Z, P, Q, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y, L)
		functions.steps(Z, P, Q, Q_tmp, k_vec, c_vec, y, L)
	else:
		functions.proSteps(Z, P, Q, Q_tmp, k_vec, c_vec)
		functions.proQuasi(Z, P, Q, Q_tmp, Q1, Q2, k_vec, c_vec)
		functions.proSteps(Z, P, Q, Q_tmp, k_vec, c_vec)
	print(f"\rPriming iteration.\t\t({time() - ts:.1f}s)\n")

	# Set up mini-batch parameters
	if args.batches > 1:
		L_bat = L_pre
		s_win = np.arange(W, dtype=np.uint32)
		W_bat = W//args.batches
		print("Mini-batch accelerated EM algorithm.")
		print(f"Using {args.batches} mini-batches.")
	else:
		print("Standard accelerated EM algorithm.")

	# Accelerated EM algorithm (fastmixture)
	ts = time()
	for it in range(args.iter):
		if args.batches > 1: # Quasi-Newton mini-batch updates
			rng.shuffle(s_win) # Shuffle window order
			for b in np.arange(args.batches):
				s_beg = b*W_bat
				s_end = W if b == (args.batches - 1) else (b + 1)*W_bat
				s_bat = s_win[s_beg:s_end]
				if args.projection is None:
					functions.batQuasi(Z, P, Q, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, s_bat, y, L)
				else:
					functions.proBatch(Z, P, Q, Q_tmp, Q1, Q2, k_vec, c_vec, s_bat)
			if args.projection is None:
				functions.quasi(Z, P, Q, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y, L)
			else:
				functions.proQuasi(Z, P, Q, Q_tmp, Q1, Q2, k_vec, c_vec)
		else: # Full updates
			if args.projection is None:
				functions.quasi(Z, P, Q, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y, L)
				functions.steps(Z, P, Q, Q_tmp, k_vec, c_vec, y, L)
			else:
				functions.proSteps(Z, P, Q, Q_tmp, k_vec, c_vec)
				functions.proQuasi(Z, P, Q, Q_tmp, Q1, Q2, k_vec, c_vec)

		# Log-likelihood convergence check
		if ((it + 1) % args.check) == 0:
			L_cur = admix_cy.loglike(Z, P, Q, c_vec)
			print(f"({it + 1})\tLog-like: {L_cur*L_nrm:.1f}\t({time() - ts:.1f}s)", flush=True)
			if args.batches > 1:
				if L_cur < (L_bat + args.tole):
					args.batches = args.batches//2 # Halve number of batches
					if args.batches > 1:
						print(f"Halving mini-batches to {args.batches}.")
						L_bat = float('-inf')
						W_bat = W//args.batches
					else: # Turn off mini-batch acceleration
						print("Running standard updates.")
					L_pre = L_cur
					if args.projection is None:
						functions.quasi(Z, P, Q, Q_tmp, P1, P2, Q1, Q2, k_vec, c_vec, y, L)
					else:
						functions.proQuasi(Z, P, Q, Q_tmp, Q1, Q2, k_vec, c_vec)
				else:
					L_bat = L_cur
			else: # Check for convergence
				if L_cur < (L_pre + args.tole):
					print("Converged!\n")
					break
				L_pre = L_cur
			ts = time()
	L_cur *= L_nrm
	print(f"Final log-likelihood: {L_cur:.1f}")

	# Save output
	Q_fin = np.zeros((N, args.K)) # Final ancestry proportions
	admix_cy.convertQ(Q, Q_fin) # Average contribution from two haplotypes
	np.savetxt(f"{args.out}.K{args.K}.s{args.seed}.Q", Q_fin, fmt="%.6f")
	print(f"Saved Q matrix as {args.out}.K{args.K}.s{args.seed}.Q")
	if not args.no_freqs and (args.projection is None):
		if F > 1: # Save P file for each file (chromosome)
			for p in np.arange(F):
				P_tmp = P[c_vec[f_vec[p]]:c_vec[f_vec[p + 1]]]
				P_tmp.tofile(f"{args.out}.K{args.K}.s{args.seed}.{args.prefix}{p + 1}.P.bin")
			with open(f"{args.out}.K{args.K}.s{args.seed}.pfilelist", "w") as f:
				for p in np.arange(F):
					f.write(f"{args.out}.K{args.K}.s{args.seed}.{args.prefix}{p + 1}.P.bin\n")
			print(f"Saved P matrices (binary) as {args.out}.K{args.K}.s{args.seed}.{args.prefix}{{1..{F}}}.P.bin")
			print(f"Saved P matrix filelist as {args.out}.K{args.K}.s{args.seed}.pfilelist")
		else: # Single file (chromosome)
			P.tofile(f"{args.out}.K{args.K}.s{args.seed}.P.bin")
			print(f"Saved P matrix (binary) as {args.out}.K{args.K}.s{args.seed}.P.bin")

	# Print elapsed time for computation
	t_tot = time() - start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"\nTotal elapsed time: {t_min}m{t_sec}s")

	# Write to log-file
	with open(f"{args.out}.K{args.K}.s{args.seed}.log", "a") as log:
		log.write(f"\nFinal log-likelihood: {L_cur:.1f}\n")
		log.write(f"Converged in {it + 1} iterations.\n")
		log.write(f"\nSaved Q matrix as {args.out}.K{args.K}.s{args.seed}.Q\n")
		if not args.no_freqs and (args.projection is None):
			if F > 1:
				log.write("Saved P matrices (binary) as " + \
			  		f"{args.out}.K{args.K}.s{args.seed}.{args.prefix}{{1..{F}}}.P.bin\n")
				log.write(f"Saved P matrix filelist as {args.out}.K{args.K}.s{args.seed}.pfilelist")
			else:
				log.write(f"Saved P matrix (binary) as {args.out}.K{args.K}.s{args.seed}.P.bin\n")
		log.write(f"\nTotal elapsed time: {t_min}m{t_sec}s\n")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'hapla admix' command!"



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
cimport numpy as np
cimport openmp as omp
from cython.parallel import parallel, prange
from libc.math cimport fmaxf, fminf, log
from libc.stdint cimport uint8_t, uint32_t
from libc.stdlib cimport calloc, free

ctypedef uint8_t u8
ctypedef uint32_t u32
ctypedef float f32

cdef f32 PRO_MIN = 1e-5
cdef f32 PRO_MAX = 1.0-(1e-5)


##### hapla - haplotype clustering #####
### Inline functions
# Inner function for marginal medians and variance
cdef inline void _marginal(
		u8* r, u32* c, const u32 n, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
	for j in range(M):
		r[j] = 1 if c[j] > n else 0
		c[j] = 0

# Calculate Hamming distance
cdef inline u32 _hammingDist(
		const u8* h, const u8* r, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
		u32 dist = 0
	for j in range(M):
		dist += 1 if h[j] != r[j] else 0
	return dist

# Calculate Hamming distance and change assignments
cdef inline u32 _hammingCheck(
		const u8* z_vec, u8* z_pre, const Py_ssize_t U
	) noexcept nogil:
	cdef:
		size_t i
		u8 z1, z2
		u32 dist = 0
	for i in range(U):
		z1 = z_pre[i]
		z2 = z_vec[i]
		z_pre[i] = z2 if z1 != z2 else z1
		dist += 1 if z1 != z2 else 0
	return dist

# Add haplotype contribution to frequency vector
cdef inline void _addHaplo(
		const u8* h, u32* c, const u32 u, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
	for j in range(M):
		c[j] += u if h[j] == 1 else 0

# Update cluster information
cdef inline void _updateClust(
		const u8* X, u8* R, u32* A, u32* B, const u32 u, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
		u8 x
	for j in range(M):
		x = X[j]
		R[j] = x
		A[j] = u if x == 1 else 0
		B[j] -= u if x == 1 else 0

# Estimate log-likelihood between cluster medians based on frequencies
cdef inline f32 _logLike(
		const u8* r, const u32* c, const f32 n, const Py_ssize_t M
	) noexcept nogil:
	cdef:
		size_t j
		f32 s = 0.0
		f32 d, f, p
	for j in range(M):
		d = <f32>r[j]
		f = <f32>c[j]*n
		p = PRO_MIN if f < PRO_MIN else (PRO_MAX if f > PRO_MAX else f)
		s += d*log(p) + (1.0 - d)*log(1.0 - p)
	return s


### Standard functions
# Compute marginal medians and cluster variances
cpdef void marginalMedians(
		u8[:,::1] R, u32[:,::1] C, const u32[::1] n_vec, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = R.shape[1]
		size_t k
	for k in range(K):
		if n_vec[k] > 0:
			_marginal(&R[k,0], &C[k,0], n_vec[k]/2, M)

# Compute distances and perform cluster assignment
cpdef void assignClust(
		u8[:,::1] X, const u8[:,::1] R, u32[:,::1] C, u8[::1] z_vec, u32[::1] c_vec, const u32[::1] n_vec, 
		u32[::1] n_tmp, const u32[::1] u_vec, const Py_ssize_t U, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[1]
		size_t i, k, x, y, z
		u8* h
		u32 c, d, u
		u32* n_thr
		u32* C_thr
		omp.omp_lock_t mutex
	omp.omp_init_lock(&mutex)
	with nogil, parallel():
		n_thr = <u32*>calloc(K, sizeof(u32))
		C_thr = <u32*>calloc(K*M, sizeof(u32))
		for i in prange(U, schedule='guided'):
			h = &X[i,0]
			z = 0
			c = M + 1
			for k in range(K):
				if n_vec[k] > 0:
					d = _hammingDist(h, &R[k,0], M)
					if d <= c:
						z = k
						c = d
			z_vec[i] = z
			c_vec[i] = c

			# Add individual contributions to temporary arrays
			u = u_vec[i]
			n_thr[z] += u
			_addHaplo(h, &C_thr[z*M], u, M)
		
		# omp critical
		omp.omp_set_lock(&mutex)
		for x in range(K):
			if n_vec[x] > 0:
				n_tmp[x] += n_thr[x]
				for y in range(M):
					C[x,y] += C_thr[x*M + y]
		omp.omp_unset_lock(&mutex)
		free(n_thr)
		free(C_thr)
	omp.omp_destroy_lock(&mutex)

# Copy and reset size of clusters
cpdef void updateN(
		u32[::1] n_vec, u32[::1] n_tmp, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
	for k in range(K):
		n_vec[k] = n_tmp[k]
		n_tmp[k] = 0

# Check and generate new cluster
cpdef u32 checkClust(
		const u8[:,::1] X, u8[:,::1] R, u32[:,::1] C, u8[::1] z_vec, const u32[::1] c_vec, u32[::1] n_vec, 
		const u32[::1] u_vec, const u32 c_lim, const Py_ssize_t U, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[1]
		Py_ssize_t L = R.shape[0]
		size_t c_arg = 0
		size_t i, z
		u32 c_max = c_vec[0]
		u32 u
	for i in range(1, U): # Find extreme point
		if c_vec[i] > c_max:
			c_arg = i
			c_max = c_vec[i]
	if c_max >= c_lim and K < L:
		u = u_vec[c_arg]
		z = z_vec[c_arg]
		_updateClust(&X[c_arg,0], &R[K,0], &C[K,0], &C[z,0], u, M)
		n_vec[K] = u
		n_vec[z] -= u
		z_vec[c_arg] = K
		return 1
	else:
		return 0

# Generate new cluster from no check
cpdef void genClust(
		const u8[:,::1] X, u8[:,::1] R, u32[:,::1] C, u8[::1] z_vec, const u32[::1] c_vec, u32[::1] n_vec, 
		const u32[::1] u_vec, const Py_ssize_t U, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[1]
		size_t c_arg = 0
		size_t i, z
		u32 c_max = c_vec[0]
		u32 u
	for i in range(1, U): # Find extreme point
		if c_vec[i] > c_max:
			c_arg = i
			c_max = c_vec[i]
	u = u_vec[c_arg]
	z = z_vec[c_arg]
	_updateClust(&X[c_arg,0], &R[K,0], &C[K,0], &C[z,0], u, M)
	n_vec[K] = u
	n_vec[z] -= u
	z_vec[c_arg] = K
	
# Convergence check through hamming distance
cpdef u32 countDist(
		const u8[::1] z_vec, u8[::1] z_tmp, const Py_ssize_t U
	) noexcept nogil:
	return _hammingCheck(&z_vec[0], &z_tmp[0], U)

# Find non-zero cluster with least assignments
cpdef u32 findZero(
		u32[::1] n_vec, const u32 N, const u32 mac, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		u32 minI = K - 1
		u32 minN = N + 1
	for k in range(K, -1, -1):
		if n_vec[k] > 0 and n_vec[k] < minN:
			minI = k
			minN = n_vec[k]
			if minN == 1:
				break
	if minN < mac:
		n_vec[minI] = 0
	return minN

# Fix index of medians
cpdef void medianFix(
		u8[:,::1] R, u32[:,::1] C, u8[::1] z_vec, u32[::1] n_vec, const Py_ssize_t K, const Py_ssize_t U
	) noexcept nogil:
	cdef:
		Py_ssize_t M = R.shape[1]
		size_t c = 0
		size_t i, j, k
		u8 z
	for k in range(K):
		if n_vec[k] > 0:
			if k != c:
				for j in range(M):
					R[c,j] = R[k,j]
					C[c,j] = C[k,j]
				for i in range(U):
					z = z_vec[i]
					z_vec[i] = c if z == k else z
				n_vec[c] = n_vec[k]
				n_vec[k] = 0
			c += 1

# Fix haplotype cluster assignments
cpdef void assignFix(
		u8[:,::1] Z, const u8[::1] z_vec, const u32[::1] p_vec, const u32[::1] d_vec, const size_t w
	) noexcept nogil:
	cdef:
		Py_ssize_t N = Z.shape[1]
		size_t u = 0
		size_t i
		u8 z = z_vec[0]
	for i in range(N):
		if d_vec[i] != 0:
			z = z_vec[u]
			u += 1
		Z[w,p_vec[i]] = z

# Reset arrays for next iteration
cpdef void resetArrays(
		u32[::1] c_vec, u32[::1] n_vec, u32[::1] p_vec, u32[::1] d_vec, u32[::1] u_vec
	) noexcept nogil:
	cdef:
		Py_ssize_t N = c_vec.shape[0]
		Py_ssize_t K = n_vec.shape[0]
		size_t i, k
	for k in range(K):
		n_vec[k] = 0
	for i in range(N):
		c_vec[i] = 0
		p_vec[i] = i
		d_vec[i] = 0
		u_vec[i] = 0

# Estimate log-likelihoods between cluster medians
cpdef void estimateLoglike(
		u8[:,::1] R, u32[:,::1] C, f32[:,::1] L, u32[::1] n_vec, Py_ssize_t K
	) noexcept nogil:
	cdef:
		Py_ssize_t M = R.shape[1]
		size_t k1, k2
	for k1 in range(K):
		for k2 in range(K):
			L[k1,k2] = _logLike(&R[k1,0], &C[k2,0], 1.0/<f32>n_vec[k2], M)
